# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from base64 import b64encode
from collections import defaultdict

from django.contrib.auth.models import User
from django.core.exceptions import ObjectDoesNotExist, PermissionDenied
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.views.decorators.http import require_GET

from pstorelib.server import urlunquote

from pstore.decorators import audit_view, nonce_required
from pstore.http import HttpError
from pstore.models import Object, ObjectPerm, PublicKey, Property
from pstore.security import get_object_or_403
from pstore.xhr import JsonResponse


@nonce_required
@require_GET
@audit_view('reads object details')
def get_object(request, object_identifier):
    """
    Get a single object. You get verbose info.

    Note that if you want info about the properties for a different user,
    you'll need to redo the query with that user id.
    """
    # Decode object_identifier.
    object_identifier = urlunquote(object_identifier)

    # Check authorization and existence:
    u = request.GET.get('u', None)
    if request.user.has_perm('object.view_any_object'):
        obj = get_object_or_404(Object, identifier=object_identifier)
    elif u != request.user.username:
        raise PermissionDenied()
    else:
        obj = get_object_or_403(Object, identifier=object_identifier)

    # Check if the user is allowed to view this.
    try:
        obj.allowed.get(user=request.user)
    except ObjectDoesNotExist:
        if not request.user.has_perm('object.view_any_object'):
            raise PermissionDenied()

    # Get lots of info for this object.
    result = {}

    # Get the allowed users.
    allowed = obj.allowed.select_related('user')
    result['users'] = dict([(i.user.username, {'can_write': i.can_write})
                            for i in allowed])

    # Get a list of properties.
    propqs = Property.objects.filter(object=obj)
    if u:
        propqs = propqs.filter(Q(user__username=u) | Q(user=None))
    else:
        propqs = propqs.filter(user=None)
    # BEWARE: For the SQLite3 backend, the LENGTH() is incorrect!
    propqs = propqs.extra(select={'size': 'LENGTH(value)'})
    properties = set(propqs.values_list('id', 'name', 'type', 'size', 'user'))

    # Get the values, but only for small properties.
    small_property_ids = [i[0] for i in properties if i[3] <= 2048]

    # https://code.djangoproject.com/ticket/9619
    # We cannot do values_list when using SQLite3.
    # #property_values = dict(Property.objects
    # #                       .filter(id__in=small_property_ids)
    # #                       .values_list('id', 'value'))
    property_qs = Property.objects.filter(id__in=small_property_ids)
    property_values = dict((i.id, i.value)
                           for i in property_qs.only('id', 'value'))

    # Put properties in the results dictionary.
    result['properties'] = {}
    enctype, encuid = None, None  # cache encryption type

    for property_id, name, type, size, user in properties:
        assert type in (Property.TYPE_PUBLIC, Property.TYPE_SHARED)

        # Cache the enctype so we don't have to look it up for every property.
        if user:
            if not enctype:
                encuid = user
                enctype = PublicKey.objects.get(user__id=user).key_type()
                assert enctype
            else:
                assert encuid == user, '%r == %r' % (encuid, user)

        # These are always returned.
        info = {'data': None,
                'size': size,
                'enctype': ('none', enctype)[bool(user)]}

        # If data is small enough, it is set too.
        if property_id in property_values:
            info['data'] = b64encode(property_values[property_id])

        result['properties'][name] = info

    # If a superuser called and wants to know what properties there are, add
    # those.
    if not u:
        # Fetch property names for the properties.
        for name, type in (Property.objects.filter(object=obj)
                           .exclude(user=None)
                           .values_list('name', 'type').distinct()):
            assert type == Property.TYPE_SHARED

            # These are always returned, but this time we have no useful info
            # for the caller.
            info = {'data': None, 'size': None, 'enctype': None}

            result['properties'][name] = info

    return JsonResponse(request, result)


@nonce_required
@require_GET
@audit_view('lists objects+info')
def list_objects(request):
    # Check authorization and existence:
    u = request.GET.get('u', None)
    if request.user.has_perm('object.view_any_object'):
        pass
    elif u != request.user.username:
        raise PermissionDenied()

    # Query strings:
    q = request.GET.get('q', None)  # filter against identifier
    v = request.GET.get('v', None) and True  # list more info

    # Query:
    qs = Object.objects.all()  # no ordering here, that's up to the client
    if q:
        qs = qs.filter(identifier__icontains=q)
    if u:
        qs = qs.filter(allowed__user__username=u)

    # Result:
    if v:
        # Listify.
        qs = list(qs.values_list('id', 'identifier'))
        # Fetch the related allow lists in a single query.
        allowed = ObjectPerm.objects.filter(object__id__in=[i[0] for i in qs])
        allowed = list(allowed.values_list('object', 'user', 'can_write'))
        users = User.objects.filter(id__in=set([i[1] for i in allowed]))
        # Dictify.
        users = dict(users.values_list('id', 'username'))
        allowed_dict = defaultdict(dict)
        for object_id, user_id, can_write in allowed:
            allowed_dict[object_id][users[user_id]] = {'can_write': can_write}
        # Combine.
        result = {}
        for object_id, identifier in qs:
            # Only add the user allow-list to the all-objects listing. If you
            # want property details, use single-object lookups.
            result[identifier] = {'users': allowed_dict[object_id]}
    else:
        result = dict((i, True)
                      for i in qs.values_list('identifier', flat=True))

    return JsonResponse(request, result)


@nonce_required
@require_GET
@audit_view('search properties')
def search_properties(request):
    # Check authorization and existence:
    u = request.GET.get('u', None)
    if request.user.has_perm('object.view_any_object'):
        pass
    elif u != request.user.username:
        raise PermissionDenied()

    # Query:
    qs = Property.objects.all()  # no ordering here, that's up to the client
    if u:
        qs = qs.filter(object__allowed__user__username=u)

    # More queries (observe that "icontains" search depends on the DB
    # backend to work insensitively; MySQL blob does *not* work).
    if 'propkey_icontains' in request.GET:
        qs = qs.filter(name__icontains=request.GET['propkey_icontains'])
    if 'propvalue_icontains' in request.GET:
        # Yuck. That length should've been memoized.
        qs = (qs
              .filter(type=Property.TYPE_PUBLIC,
                      value__icontains=request.GET['propvalue_icontains'])
              .extra(where=['LENGTH(value) <= 2048']))

    # First sanity check:
    if qs.count() > 100:
        # RFC4918 11.2.  422 Unprocessable Entity
        # > ... For example, this error condition may occur if an XML
        # > request body contains well-formed (i.e., syntactically
        # > correct), but semantically erroneous, XML instructions.
        # In this case, semantics being: be more specific.
        raise HttpError(422, 'too many results')

    # Fetch extra properties we need:
    qs = qs.extra(select={'size': 'LENGTH(value)'})
    qs = qs.select_related('object')
    qs = qs.values_list('id', 'object__identifier', 'name', 'type', 'size')

    machines = {}
    for id_, identifier, propkey, proptype, size in qs:
        propvalue = None
        if proptype == Property.TYPE_PUBLIC and size <= 2048:
            propvalue = (Property.objects.filter(id=id_)
                         .values_list('value', flat=True)[0])

        info = {
            'data': propvalue,
            'size': size,
            'enctype': ('none', 'any')[proptype != Property.TYPE_PUBLIC],
        }
        if propvalue:
            info['data'] = b64encode(propvalue)

        if identifier not in machines:
            machines[identifier] = {'properties': {}}
        machines[identifier]['properties'][propkey] = info

    return JsonResponse(request, machines)


@require_GET
@nonce_required
@audit_view('lists users+info')
def list_users(request):
    # Query strings:
    q = request.GET.getlist('q')

    # Query:
    qs = PublicKey.objects.filter(user__is_active=True).select_related('user')
    if q:
        filter = Q()
        for username in q:
            filter |= Q(user__username=username)
        qs = qs.filter(filter)

    result = {}
    for object in qs:
        result[object.user.username] = {
            'key': object.key,
            'description': object.description,
        }

    return JsonResponse(request, result)


@nonce_required
@require_GET
@audit_view('validates db consistency')
def validate(request):
    if not request.user.has_perm('object.view_any_object'):
        raise PermissionDenied()

    tpub, tshar = Property.TYPE_PUBLIC, Property.TYPE_SHARED

    errors = []

    # Assert that all objects have an allowed list.
    for object in Object.objects.filter(allowed=None):
        errors.append((
            'Object', object.id,
            '%s has no allowed list' % (object.identifier,),
            'Add ObjectPerm items',
        ))

    # FIXME: code below is broken for objects with both readers and writers..
    # # Assert that all objects have someone who can write.
    # #for object in Object.objects.filter(allowed__can_write=False):
    # #   errors.append((
    # #       'Object', object.id,
    # #       '%s does not have an admin' % (object.identifier,),
    # #       'Add ObjectPerm items with can_write powers',
    # #   ))

    # Check property types.
    for property in (Property.objects.exclude(type__in=[tpub, tshar])
                     .defer('value')):
        errors.append((
            'Property', property.id,
            ('%s -> %s has type %d' %
             (property.object.identifier, property.name, property.type)),
            'Alter the type to %d or %d' % (tpub, tshar),
        ))

    # Check property types (PUBLIC must not have a user).
    for property in (Property.objects.filter(type=tpub, user__isnull=False)
                     .defer('value')):
        errors.append((
            'Property', property.id,
            ('%s -> %s (type %d) has a user' %
             (property.object.identifier, property.name, tpub)),
            'Type %d properties should not have a user' % (tpub,),
        ))

    # Check property types (SHARED must have a user).
    for property in (Property.objects.filter(type=tshar, user__isnull=True)
                     .defer('value')):
        errors.append((
            'Property', property.id,
            ('%s -> %s (type %d) does not have a user' %
             (property.object.identifier, property.name, tshar)),
            'Type %d properties should have a user' % (tshar,),
        ))

    # Check that all type SHARED objects have as many records as there are
    # allowed users.
    for object in Object.objects.all():
        user_ids = [i.user_id for i in object.allowed.all()]
        shared_properties = (Property.objects.filter(object=object, type=tshar)
                             .defer('value').order_by('name'))

        per_property = defaultdict(list)
        for property in shared_properties:
            per_property[property.name].append(property.user_id)

        for property_name, property_user_ids in per_property.items():
            if (len(property_user_ids) != len(user_ids) or
                    set(property_user_ids) != set(user_ids)):
                errors.append((
                    'Object', object.id,
                    ('%s -> %s has wrong number of properties' %
                     (object.identifier, property_name)),
                    ('The allowed users is out of sync with the amount of '
                     'properties for this object.'),
                ))

    return JsonResponse(request, {'errors': errors})



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2010,2012,2013,2015,2017  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from datetime import datetime
from random import getrandbits

from django.contrib.auth.models import User
from django.core.validators import RegexValidator
from django.db import models
from django.db.models.signals import pre_delete
from django.utils.translation import ugettext_lazy as _

from pstorelib.crypt import encrypts
from pstorelib.exceptions import CryptError

from pstore.db import Model, AsciiField, BlobField
from pstore.notify import (notify_object_deletion, notify_publickey_change,
                           notify_user_deletion)

ascii_validator = RegexValidator(
    regex='^[\t\r\n\x20-\x7e]*$',
    message=_('Only readable ASCII is allowed: \\x20..\\x7E TAB CR LF'))


class PublicKey(Model):
    """
    The user's public key. To reduce complexity, a user is only allowed to have
    one public key. If you want multiple public keys, you should create
    multiple users.

    The key property should hold the PGP public key block:

        -----BEGIN PGP PUBLIC KEY BLOCK-----
        Version: GnuPG v1.4.11 (GNU/Linux)
        [SP]
        mI0EULkrQAEEAKU++49M+QfiSTFJjWQ8Yyr+OKa0V90aNGbYNaGvfzlPVHNS+AwR
        ...

    or the obsolete ssh public key:

        ssh-rsa AAAAB3NzaC1yc2EAAAABIwAAAIEA0IcrNyQLdVyCgpQtMfv/...

    Encrypted Property objects are encrypted using this public key.
    """
    user = models.OneToOneField(User, related_name='publickey')
    key = models.TextField(
        blank=False, validators=[ascii_validator],
        help_text=_('The user\'s public key; can be in ssh authorized_key '
                    'format (OLD) or in PGP PUBLIC KEY format (NEW).'))
    description = models.CharField(
        max_length=255, blank=True,
        help_text=_('Human readable info about the key, e.g. the PGP key '
                    'uid (Alex Boonstra (TEST) <alex@example.com>).'))

    def __init__(self, *args, **kwargs):
        super(PublicKey, self).__init__(*args, **kwargs)

        # Used for the on-public-key-change signal. (Not setting pk, so no
        # recursion.)
        if self.pk:
            self.original = PublicKey(user=self.user, key=self.key,
                                      description=self.description)

    def diff(self, other):
        """
        Used by the changes notification mechanism. Will return only those
        properties that have changed between this version and ``other``. The
        code lives here because here we always know which properties we have.
        """
        changes = {}
        for property in ('id', 'user', 'key', 'description'):
            old = getattr(self, property)
            new = getattr(other, property)
            if old != new:
                changes[property] = {'old': old, 'new': new}
        return changes

    def key_type(self):
        if self.key.startswith('ssh-rsa '):
            return 'sshrsa'
        elif self.key.startswith('----'):
            return 'gpg'
        return 'unknown'

    def save(self, **kwargs):
        # Don't save useless CRs if we don't have to, and make sure we strip
        # excess LFs and space so the key_type() function works.
        self.key = self.key.strip().replace('\r', '')
        self.description = self.description.strip().replace('\r', '')

        ret = super(PublicKey, self).save(**kwargs)
        if hasattr(self, 'original'):
            notify_publickey_change(self.original, self)

        return ret

    def __unicode__(self):
        return _(u"%(user)s's %(key_type)s public key") % {
            'user': self.user, 'key_type': self.key_type()
        }


class Object(Model):
    """
    The general object identifier with permissions.

    Generally, when you want to store the password to a machine, you'll create
    a new Object which identifies the machine:

        machine = Object.objects.create(identifier='server.example.com')

    Next, you'll attach permissions to it:

        ObjectPerm.objects.create(object=machine, user=alex, can_write=True)
        ObjectPerm.objects.create(object=machine, user=harm, can_write=True)
        ObjectPerm.objects.create(object=machine, user=walter, can_write=False)

    Attach properties to it:

        # Public properties
        PubProp(machine, 'ssh-username', 'osso')
        PubProp(machine, 'ip-address', '1.2.3.4')
        # ^-- this would be a Property object with type=PUBLIC

        # Shared (encrypted) properties; there should always be one for each
        # "allowed" user.
        SharProp(machine, 'password', 'sEcReT!', walter)
        SharProp(machine, 'password', 'sEcReT!', alex)
        SharProp(machine, 'password', 'sEcReT!', harm)
        # ^-- this would be a Property object with type=SHARED

        #FUTURE## Private properties
        #FUTURE#PrivProp(machine, 'password-walter', 'wAlTeR!', walter)
        #FUTURE# ^-- this would be a Property object with type=PRIVATE

    Certain properties may be marked as "magic" by the associated tools. E.g.
    the "password" for automatic ssh login. The Object itself does not assign
    any meaning to it though.
    """
    identifier = AsciiField(max_length=255, unique=True)

    class Meta:
        permissions = (
            # Permissions to view not-owned objects (admins get this)
            ('view_any_object', _('View any object')),
        )

    def __unicode__(self):
        return self.identifier


class ObjectPerm(models.Model):
    """
    Who has read and write permissions for an Object.
    """
    object = models.ForeignKey(Object, related_name='allowed')
    user = models.ForeignKey(User)
    can_write = models.BooleanField()

    def __unicode__(self):
        dict = {'user': self.user, 'object': self.object}
        if self.can_write:
            return _('%(user)s can write %(object)s') % dict
        return _('%(user)s can read %(object)s') % dict


class Property(models.Model):
    """
    An Object can hold one or more properties. For public properties, there
    is only one record. For private/shared (encrypted) properties, there is one
    property per allowed user.

    The value is attached to the property instead to a separate model because
    less tables are nice and renaming properties is not a recommended action.

    This property points to a user instead of to a public key, since the user
    owns the property. The key is just a means of decrypting it.
    """
    BIT_PUBLIC = 0x1
    BIT_ENCRYPTED = 0x2

    TYPE_PUBLIC = BIT_PUBLIC                  # 1
    TYPE_PRIVATE = BIT_ENCRYPTED              # 2
    TYPE_SHARED = BIT_PUBLIC | BIT_ENCRYPTED  # 3

    TYPE_CHOICES = (
        (TYPE_PUBLIC, 'Public (unencrypted)'),
        # #FUTURE# (TYPE_PRIVATE, 'Private (encrypted)'),
        (TYPE_SHARED, 'Shared (encrypted)'),
    )

    created = models.DateTimeField(auto_now_add=True)
    object = models.ForeignKey(Object, related_name='properties')
    name = AsciiField(
        max_length=255,
        help_text=_('The property identifier/name.'))
    type = models.PositiveSmallIntegerField(
        choices=TYPE_CHOICES,
        default=TYPE_PUBLIC, help_text=_('Property properties ;-)'))

    # Make sure you defer('value') unless you need it!
    value = BlobField(blank=True)

    user = models.ForeignKey(User, blank=True, null=True, default=None)

    class Meta:
        unique_together = ('object', 'name', 'user')
        verbose_name = _('property')
        verbose_name_plural = _('properties')

    def enctype(self):
        """
        What kind of encryption this property is encrypted with. All properties
        for the same user are always encrypted with the same key. Keep that in
        mind when you're checking the encryption type of multiple properties.
        (Less queries is better.)
        """
        if bool(self.type & self.BIT_ENCRYPTED):
            assert self.user_id is not None
            return PublicKey.objects.get(user__id=self.user_id).key_type()
        assert self.user_id is None
        return 'none'

    def __unicode__(self):
        return _('%(object)s -> %(name)s (%(user)s)') % {
            'object': self.object, 'name': self.name, 'user': self.user
        }


class Nonce(models.Model):
    """
    To authenticate users, we use encrypted nonces. We generate a random value,
    encrypt the data using the user's public key and hand out that nonce.

    When the user wants to do any writing, we require him to send a recent
    valid nonce. If the user has a valid nonce, we'll know that he has the
    power to decode messages using his own private key: i.e. the user is
    authenticated.

    Restrictions:
    - A nonce shall be used as most once to avoid replay attacks.
    - The user can request a number of nonces, but to avoid DoSing, we should
      limit that to a finite amount. And, we should limit the usability time,
      but not too low, since an upload might take a while if we're dealing with
      large documents.
    """
    MAX_AGE = 300       # 5 minutes should be enough to upload all properties
    MIN_LENGTH = 32     # we generate nonces of 32 bytes
    MAX_LENGTH = 32     # exactly 32

    created = models.DateTimeField(auto_now_add=True)  # expiry depends on this
    user = models.ForeignKey(User)
    value = BlobField()
    encrypted = BlobField()

    @classmethod
    def is_sane(cls, value):
        return cls.MIN_LENGTH <= len(value) <= cls.MAX_LENGTH

    @classmethod
    def is_sane_b64(cls, value):
        if not isinstance(value, basestring):
            return False
        b64_min_length = int(cls.MIN_LENGTH * 4.0 / 3)
        b64_max_length = int((cls.MAX_LENGTH + 3.0) * 4 / 3)
        return b64_min_length <= len(value) <= b64_max_length

    def is_expired(self):
        diff = datetime.now() - self.created
        try:
            diff = diff.total_seconds()
        except AttributeError:  # python2.6-
            diff = diff.days * 86400.0 + diff.seconds  # ignoring mu-seconds
        return diff > self.MAX_AGE

    def generate(self):
        assert not self.value

        value = []
        bytes = 32
        random = getrandbits(bytes * 8)  # is this random enough?

        # Long-to-bytes in a little endian fashion.
        while random:
            value.append('%c' % (random & 0xff),)
            random >>= 8
        while len(value) < bytes:
            value.append('\0')  # most significant bytes 0?

        self.value = ''.join(value)
        assert self.is_sane(self.value)

    def encrypt(self):
        assert self.value
        assert self.user
        assert not self.encrypted

        publickey = self.user.publickey
        try:
            self.encrypted = encrypts(self.value, public_key=publickey.key)
        except CryptError as e:
            raise CryptError("Encrypting for '%s' failed" % (self.user,), e)

    def save(self, **kwargs):
        # Make sure we have a nonce.
        if not self.value:
            self.generate()
        # Make sure we have an encrypted version of it.
        if not self.encrypted:
            self.encrypt()

        return super(Nonce, self).save(**kwargs)


# Time for some signal handling. The publickey change notification is not
# handled through signals but in the save() method.

def on_object_delete(instance, **kwargs):
    notify_object_deletion(object=instance)
pre_delete.connect(on_object_delete, sender=Object)  # noqa


def on_user_delete(instance, **kwargs):
    # We pass the public key and the objects because if notify had to know
    # about them, we'd get a circular dependency.
    try:
        publickey = PublicKey.objects.get(user=instance)
    except PublicKey.DoesNotExist:
        publickey = None

    object_pks = (ObjectPerm.objects.filter(user=instance)
                  .values_list('object', flat=True))
    objects = Object.objects.filter(pk__in=object_pks).order_by('identifier')

    notify_user_deletion(user=instance, publickey=publickey, objects=objects)
pre_delete.connect(on_user_delete, sender=User)  # noqa



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2010,2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.

Example project settings. Use this if pstore is the only app in your
project.

DEFAULT_FROM_EMAIL, SERVER_EMAIL, MANAGERS, ADMINS DATABASES, SECRET_KEY
should be defined in your site settings.

"""
# We're in UTC+1, we speak English and we don't do any i18n.
TIME_ZONE, LANGUAGE_CODE = 'Europe/Amsterdam', 'en-us'
USE_I18N, USE_L10N, USE_TZ = False, False, False

# Currently only used for admin-media, relative to STATIC_URL: /static/admin/
STATIC_URL = '/static/'

# Generally unused, but still needed.
SITE_ID = 1

# Middleware.
MIDDLEWARE_CLASSES = (
    # #'pstore.middleware.LogSqlToConsoleMiddleware',

    # #DJANGO1.4+#'django.middleware.clickjacking.XFrameOptionsMiddleware',
    'django.middleware.common.CommonMiddleware',

    # Make sure we have a the requirements for admin work.
    'django.contrib.sessions.middleware.SessionMiddleware',     # sessions
    'django.contrib.auth.middleware.AuthenticationMiddleware',  # request.user
    'django.contrib.messages.middleware.MessageMiddleware',     # UI feedback

    # Authenticate users by nonce instead.
    'pstore.middleware.AuthenticateByNonceMiddleware',

    # Handle HttpErrors by feeding them as response.
    # TODO/FIXME: check that the transaction is still aborted
    'pstore.middleware.HttpErrorMiddleware',

    # We want operations to be atomic! But do this after the auth-nonce
    # middleware so people won't run into the lack of nonces after they abuse
    # the pstore client (resulting in 403/404s).
    'django.middleware.transaction.TransactionMiddleware',
)

# Path to our pstore urls.
ROOT_URLCONF = 'pstore.urls'

# Python dotted path to the WSGI application used by Django's runserver.
WSGI_APPLICATION = 'pstore.wsgi.application'

# The apps that this project is comprised of.
INSTALLED_APPS = (
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',
    'pstore',
)

# Logging.
from logging import Filter
from logging.handlers import SysLogHandler


# (extra LF above for PEP)
class RequireDebugFalse(Filter):
    """For compatibility with Django 1.3-"""
    def filter(self, record):
        from django.conf import settings
        return not settings.DEBUG

LOGGING = {
    # NOTE: If you are getting log messages printed to stdout/stderr, you're
    # probably looking at a python 2.6- bug where syslog messages are encoded
    # as UTF-8 with a BOM. The BOM is read as EMERG and the message is "wall"ed
    # to all logged in users.
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'require_debug_false': {
            # In Django 1.4+ we'd use django.utils.log.RequireDebugFalse.
            '()': RequireDebugFalse,
        }
    },
    'formatters': {
        'syslog': {
            'format': 'uwsgi[%(process)d]: %(name)s: %(message)s',
        },
    },
    'handlers': {
        'mail_admins': {
            'level': 'ERROR',
            'filters': ('require_debug_false',),  # don't mail if DEBUG=False
            'class': 'django.utils.log.AdminEmailHandler',
        },
        'syslog': {
            'level': 'INFO',
            'class': 'logging.handlers.SysLogHandler',
            'address': '/dev/log',  # don't forget this for sysloghandler
            'formatter': 'syslog',
            'facility': SysLogHandler.LOG_AUTH,
        },
    },
    'loggers': {
        # Put INFO or worse in syslog.
        'pstore.audit': {
            'handlers': ('syslog',),
            'level': 'INFO',
            'propagate': True,
        },
        # Mail admins on ERROR or worse (not just for django.request).
        # And also write to the same auth.log, because it clarifies that
        # the previous statement didn't complete.
        '': {
            'handlers': ('syslog', 'mail_admins'),
            'level': 'ERROR',
            'propagate': True,
        },
    },
}



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2012,2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from base64 import b64decode

from django.core.exceptions import PermissionDenied
from django.http import Http404
from django.shortcuts import get_object_or_404

from pstore.models import Nonce


def get_object_or_403(*args, **kwargs):
    """
    Act as get_object_or_404 but raise PermissionDenied when not found. We
    don't want needless disclosure.
    """
    try:
        result = get_object_or_404(*args, **kwargs)
    except Http404:
        raise PermissionDenied()
    return result


def validate_nonce_b64(nonce_b64):
    """
    Decode the nonce_b64 and find a matching validated user.
    """
    if not Nonce.is_sane_b64(nonce_b64):
        raise PermissionDenied('Is not a nonce')

    # Re-add padding.
    nonce_b64 += (2 - ((len(nonce_b64) + 2) % 3)) * '='
    try:
        nonce = b64decode(nonce_b64)
    except TypeError:
        raise PermissionDenied('Nonce decoding failed')

    try:
        nonce = Nonce.objects.select_related('user').get(value=nonce)
    except Nonce.DoesNotExist:
        raise PermissionDenied('Nonce not found')
    else:
        if nonce.is_expired():
            raise PermissionDenied('Nonce expired')

    # Who are you?
    user = nonce.user

    # Is this user valid?
    if not user.is_active:
        raise PermissionDenied('Nonce for inactive user')

    # Ok the user is validated. Delete the nonce to make sure it's not
    # reusable.
    nonce.delete()

    return user



# vim: set ts=8 sw=4 sts=4 et ai tw=79:
"""
django-pstore -- Python Protected Password Store (Django app)
Copyright (C) 2013,2015  Walter Doekes <wdoekes>, OSSO B.V.

    This application is free software; you can redistribute it and/or modify
    it under the terms of the GNU Lesser General Public License as published
    by the Free Software Foundation; either version 3 of the License, or (at
    your option) any later version.

    This application is distributed in the hope that it will be useful, but
    WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this application; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
    USA.
"""
from base64 import b64encode
from email.Utils import formataddr

from django.contrib.admin.models import DELETION, LogEntry
from django.contrib.contenttypes.models import ContentType
from django.conf import settings
from django.core.mail import send_mail
from django.db.models import Q
from django.utils.encoding import force_unicode


def collect_object_info(property_qs, user):
    MAX_PROPERTY_SIZE = 4096

    info = []
    # BEWARE: For the SQLite3 backend, the LENGTH() is incorrect!
    properties = (property_qs.filter(Q(user=None) | Q(user=user))
                  .filter(Q(user=None) | Q(user=user))
                  .defer('value')
                  .extra(select={'size': 'LENGTH(value)'})
                  .order_by('name'))

    for property in properties:
        # Only get value if size is smaller than MAX_PROPERTY_SIZE.
        readable = None
        value = None
        if property.size <= MAX_PROPERTY_SIZE:
            value = property.value

        # The readable version can be shown when:
        # (a) it isn't encrypted
        # (b) it the value is found
        # (c) the value doesn't contain binary non-utf8 data
        if property.type & property.BIT_ENCRYPTED:
            readable = '(encrypted value of size %d)' % (property.size,)
        elif value is None:
            readable = '(value of size %d)' % (property.size,)
        else:
            try:
                readable = value.decode('utf-8')
            except UnicodeDecodeError:
                readable = None
            else:
                if any(ord(i) < 0x20 and i not in '\t\r\n'
                       for i in readable):
                    readable = None
                elif any(len(i) > 119  # mail readers can cope with 120+
                         for i in readable.split('\n')):
                    readable = None
            if readable is None:
                readable = '(non-printable)'
        # Add the property to the mail
        info.append('[%s]' % (property.name,))
        info.append('  %s' % ('\n  '.join(i.rstrip()
                              for i in readable.split('\n')).rstrip(),))
        if value is not None:
            info.append('; base64 encoded data')
            encoded = b64encode(value)
            for i in range(0, len(encoded), 70):
                info.append('  %s' % (encoded[i:(i + 70)],))
        info.append('')

    return info


def notify_object_deletion(object):
    """
    Send mail to all property owners when an object is removed.

    If a rogue admin goes around deleting objects we'll attempt to rescue some
    of the data. Decoding the shared properties from this notification mail
    should be easy enough if you have the right tools and the right private
    key.
    """
    # Compile mail info.
    subject = '[pstore] deleted object %s' % (object.identifier,)
    fromaddr = settings.DEFAULT_FROM_EMAIL

    for permission in object.allowed.select_related('user').all():
        user = permission.user

        # Skip if user doesn't have e-mail.
        if not user.email:
            continue

        # Extend mail info.
        toaddr = formataddr((user.get_full_name(), user.email))
        body = ['Object %s was deleted from the pstore!\n\n'
                'Properties are listed below.\n' % (object.identifier,)]
        body.extend(collect_object_info(property_qs=object.properties,
                                        user=user))
        flatbody = '\n'.join(body)

        # Write an admin log entry for every individual user and store
        # the flatbody there. We don't want it mailed to an insecure
        # location if we can help it.
        entry = LogEntry.objects.create(
            user=user,  # NOTE: not the *deleting* user
            content_type=ContentType.objects.get_for_model(object),
            object_id=object.pk,
            object_repr=force_unicode(object),
            action_flag=DELETION,
            change_message=flatbody
        )

        flatbody = ('Object %s was deleted from the pstore!\n\n'
                    'Details about removed properties can be found in '
                    'the admin log %d.\n' % (object.identifier, entry.id))

        # Send out a mail.
        if settings.DEBUG:
            print('Sending mail to: {0}'.format(toaddr))
            print('--')
            print(flatbody)
            print('--')
        else:
            send_mail(subject, flatbody, fromaddr, (toaddr,))


def notify_publickey_change(old_publickey, new_publickey):
    """
    Send mail to ADMINS and the user when a public key is changed.

    You'll want a notification in this case, because all new shared properties
    will hold values encrypted with the new public key while the old ones are
    unaltered. I.e. manual intervention is required unless you like to see the
    pstore store go "out of sync".
    """
    user = old_publickey.user

    # Compile mail info.
    subject = '[pstore] public key changed for %s' % (user.username,)
    fromaddr = settings.DEFAULT_FROM_EMAIL
    user_toaddrs = (formataddr((user.get_full_name(), user.email)),)
    admin_toaddrs = tuple(formataddr(i) for i in settings.ADMINS)

    body = ['Public key "%s" was changed in the pstore!\n\n'
            'Beware that adding/changing new shared properties might result '
            'in encryption with a different public key.\n' % (old_publickey,)]

    changes = old_publickey.diff(new_publickey)
    del changes['id']  # old never has a pk/id

    for change in sorted(changes.keys()):
        old, new = changes[change]['old'], changes[change]['new']
        body.append('[%s]' % (change,))
        body.append('  %s' % (old.replace('\n', '\n  '),))
        body.append('; new')
        body.append('  %s' % (new.replace('\n', '\n  '),))
        body.append('')

    flatbody = '\n'.join(body)

    # Send out a mail.
    if settings.DEBUG:
        print('Sending mail to: {0}'.format(user_toaddrs + admin_toaddrs))
        print('--')
        print(flatbody)
        print('--')
    else:
        # Two separate mails. The user does not need to know who the admins
        # are.
        for toaddrs in (user_toaddrs, admin_toaddrs):
            send_mail(subject, flatbody, fromaddr, toaddrs)


def notify_user_deletion(user, publickey, objects):
    """
    Send mail to admins when a user is removed.

    When a user goes, we'll want a record of the properties that went with it.
    This one should NOT be sent to the user but to the ADMINS instead.
    """
    objects = list(objects)  # we'll be iterating over it twice

    # Compile mail info.
    subject = '[pstore] deleted user %s' % (user.username,)
    fromaddr = settings.DEFAULT_FROM_EMAIL
    admin_toaddrs = tuple(formataddr(i) for i in settings.ADMINS)

    body = ['User %s was deleted from the pstore!\n\n'
            'The user had access to the properties tied to these objects:' %
            (user.username,)]
    for object in objects:
        body.append('- %s' % (object.identifier,))
    body.append('')

    if publickey:
        body.append('\nPublic key:\n')
        body.append('[description]\n  %s\n' %
                    (publickey.description.replace('\n', '\n  '),))
        body.append('[key]\n  %s' % (publickey.key.replace('\n', '\n  '),))
        body.append('')

    for object in objects:
        body.append('\nObject %s:\n' % (object.identifier,))
        body.extend(collect_object_info(property_qs=object.properties,
                                        user=user))
    flatbody = '\n'.join(body)

    # Do we have an admin log? Then store the flatbody there. We don't
    # want it mailed to an insecure location if we can help it.
    ct = ContentType.objects.get_for_model(user)
    extra_message = ''
    try:
        entry = LogEntry.objects.get(
            content_type=ct, object_id=user.id,
            action_flag=DELETION)
    except LogEntry.MultipleObjectsReturned:
        # In case we reset the auto_increment, we may be looking at old
        # entries as well. Take the newest one.
        entry = LogEntry.objects.filter(
            content_type=ct, object_id=user.id,
            action_flag=DELETION).latest('id')
        extra_message = (
            '\nWARNING: Multiple LogEntry items found for UID %r\n' % user.pk)
    except LogEntry.DoesNotExist:
        entry = None

    if entry:
        entry.change_message += flatbody
        entry.save()
        flatbody = ('User %s was deleted from the pstore!\n\n'
                    'Details about removed objects can be found in the admin '
                    'log %d.\n%s' % (user.username, entry.id, extra_message))

    # Send out a mail.
    if settings.DEBUG:
        print('Sending mail to: {0}'.format(admin_toaddrs))
        print('--')
        print(flatbody)
        print('--')
    else:
        send_mail(subject, flatbody, fromaddr, admin_toaddrs)
